#include "IO.h"
#include "gfx.h"
#include "core.h"

extern struct video_options_set	video_options;
extern struct app_options_set	app_options;


Uint8*					keys;
SDLMod					mods;
SDLKey					current_key;
struct mouse_status		mouse;


/****************************************************************************/
/*						GENERAL IO INTERFACE (PERIPHERAL EVENTS)			*/
/****************************************************************************/


void IO_init(void)
{
	info("IO", "starting module init..");

	IO_key_init();
	IO_mouse_init();

	info("IO", "module init completed");
}

void IO_reset(void)
{
	info("IO", "starting module reset..");

	IO_mouse_reset();
	IO_key_reset();

	info("IO", "module reset completed");
}

void IO_flush(void)
{
	IO_mouse_reset();
	IO_key_reset();
	SDL_Delay(100);
	IO_key_init();
	IO_mouse_init();
}


/****************************************************************************/
/*							MOUSE IO/EVENTS MANAGEMENT						*/
/****************************************************************************/


void IO_mouse_init(void)
{
	SDL_EventState(SDL_MOUSEBUTTONUP, SDL_IGNORE);
	SDL_EventState(SDL_MOUSEBUTTONDOWN, SDL_IGNORE);
	SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
	SDL_Delay(100);
	SDL_EventState(SDL_MOUSEBUTTONUP, SDL_ENABLE);
	SDL_EventState(SDL_MOUSEBUTTONDOWN, SDL_ENABLE);
	SDL_EventState(SDL_MOUSEMOTION, SDL_ENABLE);
	SDL_ShowCursor(SDL_DISABLE);
	mouse.buttons=SDL_GetMouseState(&mouse.x, &mouse.y);
	mouse.old_x=mouse.x;
	mouse.old_y=mouse.y;
	mouse.old_buttons=0;
	mouse.button_x=-1;
	mouse.button_y=-1;
}

void IO_mouse_reset(void)
{
	SDL_EventState(SDL_MOUSEBUTTONUP, SDL_IGNORE);
	SDL_EventState(SDL_MOUSEBUTTONDOWN, SDL_IGNORE);
	SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
	IO_mouse_buttons_release();
	SDL_ShowCursor(SDL_ENABLE);
}

void IO_mouse_buttons_release(void)
{
	mouse.old_buttons=mouse.buttons;
	mouse.buttons=0;
}

void IO_mouse_get_status(SDL_MouseMotionEvent* event)
{
	mouse.old_x=mouse.x;
	mouse.old_y=mouse.y;
	mouse.x=event->x;
	mouse.y=event->y;
	mouse.buttons=event->state;
/*	mouse.buttons=SDL_GetMouseState(&mouse.x, &mouse.y); */
}

void IO_mouse_get_button_status(SDL_MouseButtonEvent* event)
{
	mouse.old_buttons=mouse.buttons;
	mouse.button_x=event->x;
	mouse.button_y=event->y;
	mouse.buttons|=SDL_BUTTON(event->button);
}

bool IO_mouse_in_box(int x, int y, int width, int height)
/* returns true if the mouse is currently in a box */
{
	return(	(mouse.x>=x*video_options.scale)
		AND (mouse.x<(x+width)*video_options.scale)
		AND (mouse.y>=y*video_options.scale)
		AND (mouse.y<(y+height)*video_options.scale) );
}

bool IO_mouse_clicked_in_box(int button_mask, int x, int y, int width, int height)
/* returns true if a button has been clicked currently in a box */
{
	return(	IO_mouse_clicked(button_mask)
		AND IO_mouse_click_in_box(x, y, width, height) );
}

bool IO_mouse_click_in_box(int x, int y, int width, int height)
/* returns true if last button-click were in a box */
{
	return(	(mouse.button_x>=x*video_options.scale)
		AND (mouse.button_x<(x+width)*video_options.scale)
		AND (mouse.button_y>=y*video_options.scale)
		AND (mouse.button_y<(y+height)*video_options.scale)) ;
}

bool IO_mouse_clicked(int buttons_mask)
/* returns true if a set of buttons was clicked */
{
	return(	mouse.buttons&buttons_mask
		AND NOT(mouse.old_buttons&NOT(buttons_mask)) );
}

bool IO_mouse_has_moved(void)
/* returns true if the mouse has been moved */
{
	return(	(mouse.x!=mouse.old_x)
		OR	(mouse.y!=mouse.old_y) );
}

void IO_mouse_set_pointer(ptr pointer)
{
	mouse.pointer=pointer;
}


/****************************************************************************/
/*						KEYBOARD IO/EVENTS MANAGEMENT						*/
/****************************************************************************/


void IO_key_init(void)
{
	info("IO", "initializing keyboard..");

	SDL_EventState(SDL_KEYDOWN, SDL_IGNORE);
	SDL_EventState(SDL_KEYUP, SDL_IGNORE);
	SDL_Delay(100);
	SDL_EnableKeyRepeat(0, 0);
	SDL_EventState(SDL_KEYDOWN, SDL_ENABLE);
	SDL_EventState(SDL_KEYUP, SDL_ENABLE);
	IO_key_release();
	IO_key_get_status();

	info("IO", "keyboard init completed");
}

void IO_key_reset(void)
{
	SDLKey k;

	info("IO", "resetting keyboard..");

	SDL_EventState(SDL_KEYDOWN, SDL_IGNORE);
	SDL_EventState(SDL_KEYUP, SDL_IGNORE);
	IO_key_release();
	if (keys)
		for (k=SDLK_FIRST; k<=SDLK_LAST; k++)
			keys[k]=SDL_RELEASED;
	mods=KMOD_NONE;
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

	info("IO", "keyboard reset completed");
}

void IO_key_release(void)
{
	current_key=SDLK_UNKNOWN;
}

void IO_key_get_status(void)
{
	keys=SDL_GetKeyState(NULL);
	mods=SDL_GetModState();
}

bool IO_complete_key_pressed(struct complete_key compkey)
{
	if ((keys[compkey.key]==SDL_PRESSED) AND (mods==compkey.mod))
	{
#ifdef DEBUG
			printf("key (%d, %x) = (%d, %d)\n", compkey.key, compkey.mod, keys[compkey.key]==SDL_PRESSED, mods&compkey.mod);
#endif
		/*
			disable key repeat (at least for the one that has been asked for):
			pressed key value is stored, pressed status is unset
		*/
		current_key=compkey.key;
		keys[compkey.key]=SDL_RELEASED;
		return(true);
	}
	return(false);
	
}

bool IO_keymod_pressed(SDLKey key, SDLMod mod)
{
	struct complete_key compkey;

	compkey.key=key;
	compkey.mod=mod;
	return(IO_complete_key_pressed(compkey));
}

bool IO_genkey_pressed(struct generic_key genkey)
/* ask for any kind of key/mod pressed */
{
	switch (genkey.type)
	{
		case KEY:
			if (keys[genkey.code]==SDL_PRESSED)
			{
				/*
					disable key repeat (at least for the one that has been asked for):
					pressed key value is stored, pressed status is unset
				*/
				current_key=(SDLKey)genkey.code;
				keys[genkey.code]=SDL_RELEASED;
				return(true);
			}
			return(false);
			break;
		case MOD:
			return(mods&genkey.code);
			break;
	}
	return(false);
}

bool IO_key_pressed(SDLKey key)
/* ask for a key pressed */
{
	struct generic_key foo;

	foo.code=key;
	foo.type=KEY;
	return(IO_genkey_pressed(foo));
}

bool IO_mod_pressed(SDLMod mod)
/* ask for a modifier key pressed */
{
	struct generic_key foo;

	foo.code=mod;
	foo.type=MOD;
	return(IO_genkey_pressed(foo));
}

void IO_key_event(void)
{
	IO_key_get_status();
	current_key=SDLK_UNKNOWN;
}


/****************************************************************************/
/*							OTHER IO/EVENTS MANAGEMENT						*/
/****************************************************************************/


void IO_unsupported_event(SDL_Event* event)
{
	event=event;	/* tricks gcc's warning */
}

void IO_flush_events(void)
{
	SDL_Event event;

	while (SDL_PollEvent(&event));
}

bool IO_wait_until_event_or_time(Uint32 milliseconds)
{
	SDL_Event	event;
	Uint32		start, ticks, pause_start;
	bool		paused;

	paused=false;
	start=SDL_GetTicks();
	pause_start=start;
	ticks=start;
	while ((ticks-start<milliseconds) OR paused)
	{
		while (SDL_PollEvent(NULL))
		{
			SDL_WaitEvent(&event);
			switch (event.type)
			{
				case SDL_KEYDOWN:
					IO_key_event();
					if (IO_keymod_pressed(SDLK_F12, KMOD_NONE))
						gfx_save_screenshot();
					else
					{
						info("core", "wait loop aborted by user (key)");
						IO_flush();
						return(true);
					}
					break;

				case SDL_MOUSEBUTTONDOWN:
					IO_mouse_get_button_status((SDL_MouseButtonEvent*)&event);
					IO_mouse_get_status((SDL_MouseMotionEvent*)&event);
					info("core", "wait loop aborted by user (mouse button)");
					IO_flush();
					return(true);
					break;

				case SDL_QUIT:
					fast_quit();
					break;

				case SDL_ACTIVEEVENT:
					if ((event.active.state&SDL_APPACTIVE) OR (app_options.freeze_when_no_focus))
					{
						if (event.active.gain)	/* TODO: check tick reset while paused */
						{
							pause_start=SDL_GetTicks();
							paused=true;
						}
						else
						{
							start+=SDL_GetTicks()-pause_start;	/* ignore paused ticks */
							paused=false;
						}
					}
					break;

				default:
				  	NOP
			}
		}
		SDL_Delay(10);
		ticks=SDL_GetTicks();
		if (ticks<=start)	/* tick reset (after about 49 days) */
			break;
	}
	return(false);
}
