#include "core.h"
#include "gfx.h"
#include "game.h"
#include "menu.h"
#include "IO.h"
#include "intro.h"
#define __USE_POSIX
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>

#ifndef PACKAGE
#error "VERSION must be defined"
#endif

#ifndef VERSION
#error "VERSION must be defined"
#endif

#ifndef VERSION_BUILD
#error "VERSION must be defined"
#endif

extern struct video_options_set	video_options;
extern struct app_options_set	app_options;
extern SDL_Surface*			screen;
extern bool					redraw;
extern struct FPS_manager	FPS_sync;
extern int					FPS;
extern struct mouse_status	mouse;
extern character_motion		keypad_set;
extern character_motion		keypad_old;


#define GHANDLER_STACK_SIZE 16

const int		game_handler_stack_size=GHANDLER_STACK_SIZE;
game_handlers	game_handler_stack[GHANDLER_STACK_SIZE];
int				game_handler_stack_pointer;	/* stack pointer */
struct list		timers_list;
bool			timers_enabled;


/****************************************************************************/
/*						APPLICATION CODE MANAGEMENT							*/
/****************************************************************************/

void commandLine_error(const str msg, const str arg)
/*
	display an error message to stdout then exit
*/
{
	char	tmp[STR_MAX];

	sprintf(tmp, msg, arg);
	strcat(tmp, "\n");
	fatal("core", tmp);
}

void missing_argument(const str arg)
/*
	a specific error message from the command-line parser
*/
{
	commandLine_error("%s misses at least argument, try --help", arg);
}

void commandLine(const int argc, const str argv[])
/*
	command-line parser, also set default values for the options
*/
{
	int		i, value;
	str		arg;
	char	tmp[STR_MAX];

#ifdef DEBUG
	app_options.verbose=true;
#else
	app_options.verbose=false;
#endif
	app_options.skip_intro=false;
	app_options.freeze_when_no_focus=false;
	app_options.nosound=false;
	video_options.width=320;
	video_options.height=200;
	video_options.depth=8;
	video_options.scale=1;
	video_options.smooth_scaling=false;
	video_options.SDL_flags=0;	/* |SDL_DOUBLEBUF|SDL_FULLSCREEN */
	video_options.FPS=20;
	
	for (i=1; i<=argc; i++)
	{
		/* TODO */
		/* --atari --amiga */
		/* --extended-scaling */
		/* --debug */
		/* --ignore-cfg */
		/* --cfg <file> */
		arg=argv[i];
		if (strcmp(arg, "--fullscreen")==0)
			video_options.SDL_flags|=SDL_FULLSCREEN;
		else
		if (strcmp(arg, "--doublebuf")==0)
			video_options.SDL_flags|=SDL_DOUBLEBUF;
		else
		if (strcmp(arg, "--nosound")==0)
			app_options.nosound=true;
		else
		if (strcmp(arg, "--skip-intro")==0)
			app_options.skip_intro=true;
		else
		if (strcmp(arg, "--check-focus")==0)
			app_options.freeze_when_no_focus=true;
		else
		if (strcmp(arg, "--verbose")==0)
			app_options.verbose=true;
		else
		if (strcmp(arg, "--smooth-scaling")==0)
			video_options.smooth_scaling=true;
		else
		if (strcmp(arg, "--scale")==0)
		{
			if ((i<argc) AND (argv[++i]))
			{
				value=atoi(argv[i]);
				if ((value>=1) AND (value<=4))
					video_options.scale=value;
				else
				{
					sprintf(tmp, "unauthorized value for scale (1<=n<=4)\n");
					fatal("core", tmp);
				}		
			}
			else
				missing_argument(arg);
		}
		else
		if (strcmp(arg, "--fps")==0)
		{
			if ((i<argc) AND (argv[++i]))
			{
				value=atoi(argv[i]);
				if ((value>=FPS_LOWER_LIMIT) AND (value<=FPS_UPPER_LIMIT))
					video_options.FPS=value;
				else
				{
					sprintf(tmp, "unauthorized value for fps (%d<=n<=%d)\n",
							FPS_LOWER_LIMIT, FPS_UPPER_LIMIT);
					fatal("core", tmp);
				}		
			}
			else
				missing_argument(arg);
		}
		else
		if (strcmp(arg, "--help")==0)
		{
			printf("switches: --fullscreen --doublebuf --scale n --smooth-scaling --check-focus --fps n --skip-intro --verbose --help --version\n");
			printf("defaults: --scale 1\n");
			exit(0);
		}
		else
		if (strcmp(arg, "--version")==0)
		{
			printf(PACKAGE " (" PRODUCT ") " VERSION " (" VERSION_BUILD ")\n");
			exit(0);
		}
		else
			commandLine_error("unrecognized switch '%s', ignored", arg);
	}
}

#define TEMPDIR "/tmp"

void core_init(void)
/*
	perform inits for the core application
*/
{
	int			ret;
	pid_t		pid;
	fileHandle	in, out;
	char		tmp[32];
	char		filename[STR_MAX];
	
	/* check if no other instance of the program is already running: */
	/* for each rep in /proc, if the rep name is numerical, check if */
	/* the exe inside links to argv[0]. only 1 -> OK */
	if (access(TEMPDIR, F_OK))
	{
		fprintf(stderr, "Cannot access " TEMPDIR);
		exit(-1);
	}
	sprintf(filename, TEMPDIR "/" PRODUCT ".pid");
	in=fopen(filename, "r");
	if (in)
	{
		fgets(tmp, 32, in);
		pid=atol(tmp);
		if (pid)
		{
			if (kill(pid, 0)==-1)
			{
				sprintf(tmp, "ignoring stale lockfile (pid %d)\n", pid);
				info("core", tmp);
			}
			else
			{
				fprintf(stderr, PRODUCT " is already running (pid %d). Kill process or remove %s.\n",
								pid, filename);
				exit(-1);
			}
		}
	    fclose(in);
    }
	out=fopen(filename, "w");
	if (out)
	{
    	fprintf(out, "%d\n", getpid());
    	fclose(out);
	}
	else
	{
		sprintf(tmp, "%s cannot be opened for writing. Skipping lockfile protection.\n", tmp);
		info("core", tmp);
	}

	srand(SDL_GetTicks());
	
	ret=SDL_Init(SDL_INIT_TIMER|SDL_INIT_VIDEO|(app_options.nosound?0:SDL_INIT_AUDIO));
	if (ret<0)
	{
		fprintf(stderr, "SDL init error %d\n", ret);
		exit(-1);
	}
	atexit(SDL_Quit);

	SDL_WM_SetCaption(PACKAGE, PACKAGE);

	gfx_init();
	IO_init();

	list_init(&timers_list);
	timers_lock();

	init_framerate(&FPS_sync);
	set_framerate(&FPS_sync, video_options.FPS);
}

void core_reset(void)
/*
	perform un-inits for the core application
*/
{
	char	filename[STR_MAX];

	IO_reset();
	gfx_reset();

	SDL_ShowCursor(SDL_ENABLE);

	sprintf(filename, TEMPDIR "/" PRODUCT ".pid");
	remove(filename);
}

void fast_quit(void)
/*
	roughly flush the game_handle stack (this will leave the app)
*/
{
	gfx_clear(screen);
	gfx_flush();
	while (get_current_game_handler()!=GHANDLER_ROOT)
		pop_game_handler();
	pop_game_handler();
}


/****************************************************************************/
/*							GAME HANDLER MANAGEMENT							*/
/****************************************************************************/


void push_game_handler(const game_handlers game_handler)
/*
	add a game_handler to the stack
	call the relevant *_init() stuff, may also set the redraw flag
*/
{
	char	tmp[STR_MAX];

	if (game_handler_stack_pointer>=game_handler_stack_size)
		fatal("core.GHstack", "game_handler_stack overflow!\n");
	else
	{
		game_handler_stack[game_handler_stack_pointer++]=game_handler;
		switch (game_handler)
		{
			case GHANDLER_ROOT:
				core_init();
				break;
			case GHANDLER_INTRO:
				intro_init();
				break;
			case GHANDLER_MENU:
				menu_init();
				break;
			case GHANDLER_CREDITS:
				credits_init();
				break;
			case GHANDLER_GAME:
				game_init();
				break;
			case GHANDLER_MAZE:
				maze_init();
				redraw=true;
				break;
			case GHANDLER_INVENTORY:
				inventory_init();
				redraw=true;
				break;
			case GHANDLER_FILE_MANAGER:
				break;
			case GHANDLER_OPTIONS:
				break;
			case GHANDLER_HELP:
				break;
			case GHANDLER_PAUSED:
				break;
			case GHANDLER_EPILOG:
				break;
			default:
				sprintf(tmp, "pushing unexpected game handler %d", game_handler);
				fatal("core", tmp);
		}
	}
}

void pop_game_handler(void)
/*
	remove the top game_handler from the stack
	call the relevant *_reset() stuff, may also set the redraw flag
*/
{
	char	tmp[STR_MAX];

	if (game_handler_stack_pointer>0)
	{
		/* update the stack's status */
		game_handler_stack_pointer--;

		/* remove timer events from the current game handler BEFORE calling its *reset() stuff */
		timers_flush_filtered(game_handler_stack[game_handler_stack_pointer]);

		switch (game_handler_stack[game_handler_stack_pointer])
		{
			case GHANDLER_ROOT:
				core_reset();
				exit(0);
				break;
			case GHANDLER_INTRO:
				intro_reset();
				break;
			case GHANDLER_MENU:
				menu_reset();
				break;
			case GHANDLER_CREDITS:
				credits_reset();
				break;
			case GHANDLER_GAME:
				game_reset();
				redraw=true;
				break;
			case GHANDLER_MAZE:
				maze_reset();
				break;
			case GHANDLER_INVENTORY:
				inventory_reset();
				redraw=true;
				break;
			case GHANDLER_FILE_MANAGER:
				break;
			case GHANDLER_OPTIONS:
				break;
			case GHANDLER_HELP:
				break;
			case GHANDLER_PAUSED:
				break;
			case GHANDLER_EPILOG:
				break;
			default:
				sprintf(tmp, "poping unexpected game handler %d",
						game_handler_stack[game_handler_stack_pointer]);
				fatal("core.GHstack", tmp);
		}
	}
	else
		fatal(NULL, "game_handler_stack underflow!\n");
}

game_handlers get_current_game_handler(void)
/*
	return the top game_handler in the stack
*/
{
	if ((game_handler_stack_pointer<=0)
	OR (game_handler_stack_pointer>game_handler_stack_size))
		fatal("core.GHstack", "game_handler_stack corrupted!\n");
	return(game_handler_stack[game_handler_stack_pointer-1]);
}

void reset_game_handler_stack(void)
/*
	forced reset/clearing of the game_handle stack
*/
{
	int i;

	for (i=0; i<game_handler_stack_size; i++)
		game_handler_stack[game_handler_stack_pointer]=GHANDLER_NONE;
	game_handler_stack_pointer=0;
}


/****************************************************************************/
/*							GAME TIMERS SCHEDULING							*/
/****************************************************************************/


bool timers_scheduler(void)
/*
	for each timer in the global list, check if the timer has to be
	raised (according to a number of ticks passed)
	if so, compute its data
	if the timer expires, remove it from the list
*/
{
	p_cell			c, d;
	p_timer_event	timer;
	Uint32			ticks;
	bool			ret;

	ret=false;
	if (timers_enabled)
	{
		ticks=SDL_GetTicks();
		c=timers_list.head;
		while (c)
		{
			timer=c->content;
			d=c->next;	/* immediately get the next, in case c has to be removed */
			/* check if this timer has to be raised */
			if (timer AND ((ticks-timer->ticks)>=timer->ticks_period))
			{
				/* perform the actions raised by this timer */
				timers_compute_data(timer);
				ret=true;

				timer->elapsed_periods++;
				timer->ticks=ticks;
				/* check for timer expiration */
				if (timer->expires AND (timer->remaining!=0))
				{
					timer->remaining--;
					if (timer->remaining==0)
					{
						/* the timer has expired, remove it from the list */
						timers_remove_data((p_timer_event)c->content);
						xfree((ptr)&c->content);
						list_remove_cell(&timers_list, c);
						xfree((ptr)&c);
					}
				}
			}
			c=d;
		}
	}
	ret|=	(keypad_set!=CMOTION_STANDING)
		OR	(keypad_old!=keypad_set);
	return(ret);
}

void timers_register_new(const Uint32 period, const bool expires, const Uint32 remaining,
							const timer_type type, const ptr data, const bool exclusive)
/*
	add a new timer to the global list
*/
{
	p_timer_event	timer;
	p_cell			c, d;

	timer=(p_timer_event)malloc(sizeof(struct timer_event));
	if (timer==NULL)
		fatal("core.timers", "cannot allocate new timer_event");
	timer->game_handler=get_current_game_handler();
	timer->ticks=SDL_GetTicks();
	timer->ticks_period=period;
	timer->expires=expires;
	timer->remaining=remaining;
	timer->elapsed_periods=0;
	timer->type=type;
	timer->data=data;
	/*
		if this timer is exclusive for its data, remove all the previously
		registered timers that use this data
	 */
	if (exclusive)
	{
		c=timers_list.head;
		while (c)
		{
			d=c->next;		/* immediately get the next, in case c has to be removed */
			if (c->content AND ((p_timer_event)c->content)->data==data)
			{
				timers_remove_data((p_timer_event)c->content);
				xfree((ptr)&c->content);
				list_remove_cell(&timers_list, c);
				xfree((ptr)&c);
#ifdef DEBUG
				info("core.timers", "removed concurrent timer");
#endif
			}
			c=d;
		}
	}
	list_push(&timers_list, (ptr)timer);
#ifdef DEBUG
	info("core.timers", "new timer registered");
#endif
}

void timers_compute_data(p_timer_event timer)
/*
	perform the actions raised by a timer (which is determined
	following the timer type)
*/
{
	p_map_door	door;

#ifdef DEBUG
	info("core.timers", "timer processing");
#endif
	/* TODO: manage timer's data according to the timer type */
	switch (timer->type)
	{
		case TIMER_DOOR:
			door=(p_map_door)timer->data;
			/* check collision w/ team or character */
			if (door->motion==DMOTION_CLOSING)
			{
				if (door->stage>DOOR_CLOSED)
					door->stage--;
			}
			else
			{
				if (door->stage<DOOR_OPENED)
					door->stage++;
			}
			if ((door->stage==DOOR_CLOSED) OR (door->stage==DOOR_OPENED))
			{
				door->motion=DMOTION_NONE;
				timer->expires=true;
				timer->remaining=1;
			}
			redraw=true;
			break;
		default:
			NOP
	}
}

void timers_remove_data(p_timer_event timer)
/*
	free a timer's data, according to the data content (which is determined
	following the timer type and is optional)
*/
{
#ifdef DEBUG
	info("core.timers", "timer expiration");
#endif
	/* TODO: free timer's data according to the timer type */
	switch (timer->type)
	{
		case TIMER_DOOR:
			NOP
			break;
		default:
			NOP
	}
}

void timers_flush(void)
/*
	remove all the registered times from the global list
*/
{
	p_cell	c;

	c=timers_list.head;
	while (c)
	{
		if (c->content)
		{
			timers_remove_data((p_timer_event)c->content);
			xfree((ptr)&c->content);
		}
		c=c->next;
	}
	list_clear(&timers_list);
#ifdef DEBUG
	info("core.timers", "timers flushed");
#endif
}

void timers_flush_filtered(const game_handlers game_handler)
/*
	remove all the registered times of a specific game handle from the global list
*/
{
	p_cell			c, d;
	p_timer_event	timer;

	c=timers_list.head;
	while (c)
	{
		timer=c->content;
		d=c->next	/* immediately get the next, in case c has to be removed */;
		if (timer AND (timer->game_handler==game_handler))
		{
			timers_remove_data((p_timer_event)c->content);
			xfree((ptr)&c->content);
			list_remove_cell(&timers_list, c);
			xfree((ptr)&c);
		}
		c=d;
	}
#ifdef DEBUG
	info("core.timers", "timers filtered");
#endif
}

void timers_lock(void)
{
	timers_enabled=false;
}

void timers_unlock(void)
{
	timers_enabled=true;
}


/****************************************************************************/
/*								THE MAIN LOOP								*/
/****************************************************************************/


int main(int argc, str argv[])
{
	char			tmp[STR_MAX];
	SDL_Event		event;
	bool			paused, new_IO_event, new_other_event, first;

	/* app set up */
	commandLine(--argc, argv);

	reset_game_handler_stack();
	push_game_handler(GHANDLER_ROOT);

	/* intro sequence */
	if (app_options.skip_intro==false)
	{
		push_game_handler(GHANDLER_INTRO);
		intro_play();
		pop_game_handler();
	}

	/* enter the main interactive loop */
	push_game_handler(GHANDLER_MENU);
	paused=false;
	first=true;
	timers_unlock();
	while (get_current_game_handler()!=GHANDLER_ROOT)
	{
		new_IO_event=false;
		while (SDL_PollEvent(NULL))
		{
			SDL_WaitEvent(&event);
			switch (event.type)
			{
				case SDL_KEYUP:
					IO_key_release();
					break;

				case SDL_KEYDOWN:
					/* some exclusive special keys (global behaviour) */
			  	    IO_key_event();
					if (IO_keymod_pressed(SDLK_F12, KMOD_NONE))
						gfx_save_screenshot();
					else
						new_IO_event=true;
					break;

				case SDL_MOUSEBUTTONUP:
					IO_mouse_buttons_release();
					break;

				case SDL_MOUSEBUTTONDOWN:
					IO_mouse_get_button_status((SDL_MouseButtonEvent*)&event);
					new_IO_event=true;
					break;

				case SDL_MOUSEMOTION:
					IO_mouse_get_status((SDL_MouseMotionEvent*)&event);
					if (IO_mouse_has_moved())
						new_IO_event=true;
					break;

				case SDL_QUIT:
					fast_quit();
					break;

				case SDL_ACTIVEEVENT:
					if ((event.active.state&SDL_APPACTIVE) OR (app_options.freeze_when_no_focus))
					{
						if (event.active.gain)
						{
							/* TODO: release frozen timers */
							info("core", "woke up");
							paused=false;
							timers_unlock();
						}
						else
						{
							/* TODO: freeze timers */
							info("core", "paused");
							paused=true;
							timers_lock();
						}
					}
					break;

				default:
				  	IO_unsupported_event(&event);
			}
		}
		/* schedule timers (game events preemption) */
		new_other_event=timers_scheduler();
		
		/*	following the game stage, and IF necessary, call the relevant game handler
			in order to manage the IO/timer events raised */
		if (((paused==false) AND (new_IO_event OR new_other_event)) OR first)
		{
			if (get_current_game_handler()!=GHANDLER_ROOT)
			{
				switch (get_current_game_handler())
				{
					case GHANDLER_MENU:
						menu_event_handler(new_IO_event?&event:NULL);
						break;
					case GHANDLER_MAZE:
						maze_event_handler(new_IO_event?&event:NULL);
						break;
					case GHANDLER_INVENTORY:
						inventory_event_handler(new_IO_event?&event:NULL);
						break;
					case GHANDLER_OPTIONS:
						break;
					case GHANDLER_HELP:
						break;
					case GHANDLER_PAUSED:
						break;
					default:
						sprintf(tmp, "unexpected game handler %d", get_current_game_handler());
						fatal("core", tmp);
				}

				/* render the scene */
				if (get_current_game_handler()!=GHANDLER_ROOT)
					gfx_draw_screen();
			}
		}
		/* video sync (against framerate) */
		framerate_delay(&FPS_sync);
		first=false;
	}
	gfx_clear(screen);
	gfx_flush();

	/* credits sequence */
	if (app_options.skip_intro==false)
	{
		push_game_handler(GHANDLER_CREDITS);
		credits_display();
		pop_game_handler();
	}

	/* natural dying */
	pop_game_handler();
	return(0);
}
