#include "util.h"
#include "core.h"


struct app_options_set app_options;


void message(const str module, const str type, const str text)
{
	if (NOT(text) OR NOT(type))
	{
		if (module AND *module)
			fprintf(stdout, PRODUCT ":ERROR: internal error in message mgr (%p='%s',%p='%s',%p='%s')\n",
					module, module?module:"", type, type?type:"", text, text?text:"");
		else
			fprintf(stdout, PRODUCT ":ERROR: undefined error\n");
	}
	if (module AND *module)
		fprintf(stdout, PRODUCT ":%s [module=%s] %s\n", type, module, text);
	else
		fprintf(stdout, PRODUCT ":%s: %s\n", type, text);
	fflush(stdout);
}

void info(const str module, const str text)
{
	if (app_options.verbose)
		message(module, "INFO", text);
}

void warning(const str module, const str text)
{
	if (app_options.verbose)
		message(module, "WARNING", text);
}

void error(const str module, const str text)
{
	message(module, "ERROR", text);
}

void fatal(const str module, const str text)
{
	message(module, "FATAL ERROR", text);
	core_reset();
	exit(-1);
}

ptr xfree(ptr* pointer)
{
	if (pointer)
	{
		if (*pointer)
			free(*pointer);
		*pointer=NULL;
	}
	return(NULL);
}

str strdup(const str text)
{
	str tmp;

	if (text==NULL)
		return(NULL);
	tmp=malloc(strlen(text)+1);
	if (tmp==NULL)
		return(NULL);
	strcpy(tmp, text);
	return(tmp);
}

void list_insert(const p_list thelist, const ptr content)
{
	p_cell c;

	if (thelist==NULL)
		fatal("list", "attempting to insert to a NULL p_list");
	if (content==NULL)
		fatal("list", "attempting to insert a NULL content to a list");

	c=(p_cell)malloc(sizeof(struct cell));
	if (c==NULL)
		fatal("list", "cannot allocate new cell in list");
	c->content=content;
	c->previous=NULL;
	c->next=thelist->head;
	if (thelist->head==NULL)
		thelist->tail=c;
	else
		(thelist->head)->previous=c;
	thelist->head=c;
	thelist->size++;
}

void list_push(const p_list thelist, const ptr content)
{
	p_cell c;

	if (thelist==NULL)
		fatal("list", "attempting to push to a NULL p_list");
	if (content==NULL)
		fatal("list", "attempting to push a NULL content to a list");

	c=(p_cell)malloc(sizeof(struct cell));
	if (c==NULL)
		fatal("list", "cannot allocate new cell in list");
	c->content=content;
	c->previous=thelist->tail;
	c->next=NULL;
	if (thelist->tail==NULL) /* is this an empty list? (where head==tail==NULL */
		thelist->head=c;
	else
		(thelist->tail)->next=c;
	thelist->tail=c;
	thelist->size++;
}

void list_pop(const p_list thelist)
{
	p_cell c;

	if (thelist==NULL)
		fatal("list", "attempting to pop to a NULL p_list");
	c=thelist->tail;
	if (c==NULL)	/* list has a tail? (at least 1 cell) */
	{
		warning("list", "attempting to pop from an empty list");
		return;
	}

	if (c->previous)	/* test will match is more than 1 cell */
	{
		xfree((ptr)&c->content);
		c=c->previous;
		xfree((ptr)&c->next);
		thelist->tail=c;
	}
	else
	{
		xfree((ptr)&thelist->tail->content);
		xfree((ptr)&thelist->tail);
		thelist->head=NULL;
	}
	thelist->size--;
}

void list_clear(const p_list thelist)
{
	p_cell c;

	if (thelist==NULL)
		return;
/*		fatal("list", "attempting to clear a NULL list");*/

	c=thelist->head;
	while (c)
	{
		xfree((ptr)&c->content);	/* delete the cell's content */
		c=c->next;
		xfree((ptr)&c->previous);	/* delete the cell itself (indirect addressing) */
	}
	thelist->head=NULL;
	thelist->tail=NULL;
	thelist->size=0;
}

void list_init(const p_list thelist)
{
	if (thelist==NULL)
		fatal("list", "attempting to init a NULL p_list");

	thelist->head=NULL;
	thelist->tail=NULL;
	thelist->size=0;
}

void list_remove(const p_list thelist, const ptr content)
{
	p_cell c, d;
	
	if (thelist==NULL)
		fatal("list", "attempting to remove from a NULL p_list");
	if (content==NULL)
		fatal("list", "attempting to remove a NULL content from a list");

	c=thelist->head;
	while (c)
	{
		if (c->content==content)
		{
			d=c;
			c=c->next;
			list_remove_cell(thelist, d);
			xfree((ptr)d);
			return;
		}
		else
			c=c->next;
	}
	warning("list", "couldn't find a content to remove from a list");
}

void list_remove_cell(const p_list thelist, const p_cell thecell)
/* 1- thecell MUST be contained in thelist (it will be removed from this list) */
/* 2- does NOT free the cell, but its content only */
{
	if (thelist==NULL)
		fatal("list", "attempting to remove from a NULL p_list");
	if (thecell==NULL)
		fatal("list", "attempting to remove a NULL cell from a list");

	xfree((ptr)&thecell->content);
	if (thecell->previous)
		(thecell->previous)->next=thecell->next;	/* link the previous one to the next one */
	else
		thelist->head=thecell->next; /* removing the head, update the list's attribute  */
	if (thecell->next)
		(thecell->next)->previous=thecell->previous; /* link the next one to the previous one */
	else
		thelist->tail=thecell->previous;	/* removing the tail, update the list's attribute */	
	thelist->size--;
}

#ifdef DEBUG
void list_tests(void)
{
	str a, b, c, d;
	struct list mylist;
	p_cell C;

	a=strdup("hello");
	b=strdup("I'm");
	c=strdup("a little");
	d=strdup("rabbit");

	list_clear(&mylist);

	list_insert(&mylist, (ptr)a);
	list_insert(&mylist, (ptr)b);
	list_insert(&mylist, (ptr)c);
	list_insert(&mylist, (ptr)d);

	C=mylist.head;
	while (C)
	{
		printf("%s\n", (str)C->content);
		C=C->next;
	}

	C=mylist.tail;
	while (C)
	{
		list_remove_cell(&mylist, C);
		C=C->previous;
	}

	C=mylist.head;
	while (C)
	{
		printf("%s\n", (str)C->content);
		C=C->next;
	}
}
#endif

bool file_exists(const str filename)
/* returns true if the file exists (file, not folder, R/W access)
   test read AND write access rights.
*/
{
	fileHandle thefile;

	thefile=fopen(filename, "r+b");
	if (thefile)
	{
		fclose(thefile);
		return(true);
	}
	return(false);
}
